/* David Leonard, 2002. Public domain. */
/*
 * Linux cloop files are compressed filesystem images.
 * Cloop files start with a 128 byte 'gap' (filled with a shell script
 * usually), then two 32-bit network-endian integers representing the block 
 * size and the block count of the compressed filesystem. Next follows an 
 * array of 32-bit network-endian integer offsets into the source file for
 * each zlib-compressed block (or 'cblk'). The array length is the same as
 * the block count plus 1. The last offset in the array is not for a real cblk;
 * but it helps when computing the last block's size.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <zlib.h>

/* zlib error message tables from zutil.h */
extern const char *z_errmsg[10];
#define ERR_MSG(err) z_errmsg[Z_NEED_DICT-(err)]

/* cloop file header */
struct cloop {
	char		pad[128];
	u_int32_t	blocksize;
	u_int32_t	numblocks;
	u_int32_t	cblkoff[1];
};

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct stat sb;
	int fd;
	caddr_t fs, cblk;
	u_int32_t blocksize, numblocks, block;
	struct cloop *cloop;
	char *buf;
	uLongf buflen;
	int zerr, cblklen;

	/* Check usage */
	if (argc != 1) {
		fprintf(stderr, "usage: %s < fs.cloop > fs.img\n", argv[0]);
		exit(1);
	}
	fd = STDIN_FILENO;

	/* Map the cloop file into memory; possible because offsets are 32bit */
	if (fstat(fd, &sb) == -1)
		err(1, "fstat");
	if (!S_ISREG(sb.st_mode))		/* XXX necessary? */
		warnx("stdin is not a regular file");
	fs = mmap(NULL, sb.st_size, PROT_READ, MAP_FILE, fd, 0);
	if (fs == (caddr_t)MAP_FAILED)
		err(1, "mmap");
	close(fd);

	/* Extract key information */
	cloop = (struct cloop *)fs;
	blocksize = ntohl(cloop->blocksize);
	numblocks = ntohl(cloop->numblocks);

	fprintf(stderr, "extracting %u blocks, each of size %u bytes\n",
	    numblocks, blocksize);

	/* Allocate output storage for decompression */
	buf = (char *)malloc(blocksize);

	/* Decompress each block to stdout */
	for (block = 0; block < numblocks; block++) {
		cblk = fs + ntohl(cloop->cblkoff[block]);
		cblklen = ntohl(cloop->cblkoff[block+1]) -
		    ntohl(cloop->cblkoff[block]);
		buflen = blocksize;
		if ((zerr = uncompress(buf, &buflen, cblk, cblklen)) != Z_OK)
			errx(1, "block %u: uncompress: %s", block,
			    ERR_MSG(zerr));
		if (buflen != blocksize)
			warnx("block %u: short", block);
		if (write(STDOUT_FILENO, buf, blocksize) == -1)
			err(1, "write");
	}

	free(buf);
	exit(0);
}
